# -*- coding: utf-8 -*-

# NOTE: tests directory can go either inside or outside package directory.
# In cases where the production environment is likely to be quite different
# from the development environment, it can be useful to distribute the
# tests to be able to run them in the production environment.
